"""This module provides a SearchResults class."""

from types import GeneratorType
from ..resources import Record, StatsResult
from ..swimlane_dict import SwimlaneDict


__metaclass__ = type


class SearchResult:
    """A class that wraps a Swimlane search result."""

    def __init__(self, report, resp):
        """Init a SearchResult.

        Args:
            report (Report): The report that was used to initiate the search.
            resp (SwimlaneDict): The JSON response from a search request.
        """
        self.is_stats = isinstance(resp, GeneratorType)
        if self.is_stats:
            self.stats = (StatsResult(SwimlaneDict(r)) for r in resp)
        else:
            self.report = report
            self.count = resp["count"]
            self.offset = resp["offset"]
            self.limit = resp["limit"]
            results = []
            if report.applicationIds[0] in resp['results']:
                results = resp["results"][report.applicationIds[0]]
            self.records = (Record(SwimlaneDict(r)) for r in results)



import mock
import unittest

from swimlane.core.resources import Record


MOCK_RECORD = {
    'id': '123',
    'applicationId': '456',
    'values': {
        'foo': 'bar'
    }
}


class RecordTestCase(unittest.TestCase):
    def test_init(self):
        record = Record(MOCK_RECORD)
        for key, value in MOCK_RECORD.items():
            self.assertEqual(getattr(record, key), value)

    @mock.patch('swimlane.core.resources.record.Client', autospec=True)
    def test_insert(self, mock_client):
        mock_client.post.return_value = MOCK_RECORD
        record = Record(MOCK_RECORD)
        record.insert()
        mock_client.post.assert_called_once_with(record, 'app/456/record')

    @mock.patch('swimlane.core.resources.record.Client', autospec=True)
    def test_update(self, mock_client):
        mock_client.put.return_value = MOCK_RECORD
        record = Record(MOCK_RECORD)
        record.isNew = False
        record.update()
        mock_client.put.assert_called_once_with(record, 'app/456/record')

    @mock.patch('swimlane.core.resources.record.Client', autospec=True)
    def test_add_comment(self, mock_client):
        record = Record(MOCK_RECORD)
        record.add_comment('123', '456', 'Test Comment')
        mock_client.post.assert_called_once_with({
            'message': 'Test Comment',
            'createdDate': mock.ANY
        }, 'app/456/record/123/123/comment')

    @mock.patch('swimlane.core.resources.record.Client', autospec=True)
    def test_references(self, mock_client):
        record = Record(MOCK_RECORD)
        record.references('123', ['456'], ['789'])
        mock_client.get.assert_called_once_with(
            'app/456/record/123/references?recordIds=456&fieldIds=789')

    @mock.patch('swimlane.core.resources.record.Client', autospec=True)
    def test_new_for(self, mock_client):
        Record.new_for('123')
        mock_client.get.assert_called_once_with(
            'app/123/record')

    @mock.patch('swimlane.core.resources.record.Client', autospec=True)
    def test_find(self, mock_client):
        Record.find('123', '456')
        mock_client.get.assert_called_once_with('app/123/record/456')



"""This module provides a search API."""

from ..auth import Client
from ..resources import StatsReport
from .search_result import SearchResult


__metaclass__ = type


class Search:
    """A class for searching Swimlane records."""

    def __init__(self, report):
        """Init a Search.

        Args:
            report (Report): The Report or StatsReport to search with.
        """
        self.report = report
        self.has_more_pages = False

    def execute(self):
        """Execute the search.

        Returns:
            A SearchResult.
        """
        url = "search"
        if isinstance(self.report, StatsReport):
            url += "/stats"
        result = SearchResult(self.report, Client.post(self.report, url))
        if not result.is_stats:
            self.report.offset += 1
            self.has_more_pages = (result.count >
                                   self.report.offset * self.report.pageSize)
        return result

    def next_page(self):
        """Retrieve the next page of results from the search.

        Returns:
            A SearchResult.
        """
        return self.execute()



import unittest

from swimlane.core.search.filtering import create_filter

class FilteringTestCase(unittest.TestCase):
    def test_create_filter(self):
        filter = create_filter('foo', 'equals', 'bar')
        self.assertEqual(filter, {
            'fieldId': 'foo',
            'filterType': 'equals',
            'value': 'bar'})



import mock
import unittest

from swimlane.core.resources import App
from swimlane.utils import copy_field_values


class CopyFieldValuesTestCase(unittest.TestCase):
    def setUp(self):
        self.src_app = App({
            'id': '123',
            'name': 'Source',
            'acronym': 'SO',
            'fields': [{
                'name': 'Component',
                'values': [
                    {'name': 'Foo'},
                    {'name': 'Bar'},
                    {'name': 'Baz'}],
                'fieldType': 'valuesList'}]
        })
        self.dest_app = App({
            'id': '456',
            'name': 'Destination',
            'acronym': 'DE',
            'fields': [{
                'name': 'Component',
                'values': [
                    {'name': 'Baz'},
                    {'name': 'Bar'}],
                'fieldType': 'valuesList'}]
        })

    def test_copy_field_values(self):
        with mock.patch('swimlane.core.resources.App.save') as mock_save:
            added, moved = copy_field_values(
                self.src_app, 'Component',
                self.dest_app, 'Component')
        self.assertEqual(len(added), 1)
        self.assertEqual(added[0]['name'], 'Foo')
        self.assertEqual(len(moved), 1)
        self.assertEqual(moved[0]['name'], 'Bar')
        mock_save.assert_called_once()



from mock import patch
import unittest

from swimlane.core.auth.user_pass_auth_provider import UserPassAuthProvider


class UserPassAuthProviderTestCase(unittest.TestCase):
    def test_init(self):
        u = UserPassAuthProvider('server', 'user', 'pass')
        self.assertEqual(u.base_url, 'user/login')
        self.assertEqual(u.username, 'user')
        self.assertEqual(u.password, 'pass')
        self.assertEqual(u.verify_ssl, True)

    @patch('swimlane.core.auth.user_pass_auth_provider.requests', autospec=True)
    def test_auth_header(self, mock_requests):
        mock_cookie = {'.AspNetCore.Identity.Application': 'testcookie'}
        mock_requests.post.return_value.cookies = mock_cookie

        u = UserPassAuthProvider('server', 'user', 'pass')
        h = u.auth_header()

        self.assertEqual(h, {'Cookie': '.AspNetCore.Identity.Application=testcookie'})
        mock_requests.post.assert_called_once_with(
            'user/login',
            json={'username': 'user', 'password': 'pass'},
            verify=True)
        mock_requests.post.return_value.raise_for_status.assert_called_once_with()  # noqa



import mock
import unittest

from swimlane.records import add_references


class RecordsTestCase(unittest.TestCase):
    @mock.patch('swimlane.records.App', autospec=True)
    def test_add_references_app_not_found(self, mock_app):
        mock_app.find.return_value = None
        self.assertRaises(
            Exception,
            lambda: add_references('123', 'keywords'))
        mock_app.find.assert_called_once_with(app_id=None, name=None)

    @mock.patch('swimlane.records.App', autospec=True)
    def test_add_references_field_id_not_found(self, mock_app):
        mock_app_instance = mock.MagicMock()
        mock_app_instance.field_id.return_value = None
        mock_app.find.return_value = mock_app_instance
        self.assertRaises(
            Exception,
            lambda: add_references('123', 'keywords', field_name='foo'))
        mock_app_instance.field_id.assert_called_once_with('foo')



import binascii
import copy
import os
import time

from .core.resources import App


def random_objectid():
    """Returns a randomly generated MongoDB ObjectId."""
    timestamp = '{0:x}'.format(int(time.time()))
    rest = binascii.b2a_hex(os.urandom(8)).decode('ascii')
    return timestamp + rest


def scrub(obj, *keys):
    """
    Recursively remove one or more keys from a dict or list object.

    :param obj: A dict or list to scrub.
    :type obj: dict or list
    :param keys: One or more keys to remove.
    :param keys: str
    """
    if isinstance(obj, dict):
        for k in obj.keys():
            if k in keys:
                del obj[k]
            else:
                scrub(obj[k], *keys)
    elif isinstance(obj, list):
        for i in reversed(range(len(obj))):
            if obj[i] in keys:
                del obj[i]
            else:
                scrub(obj[i], *keys)
    else:
        pass


def get_by_key_value(obj, key, value, default=None):
    """
    Returns an object in a list with a matching key and value.

    :param obj: The list to search for object.
    :type obj: list
    :param key: The key to look for.
    :type key: str
    :param value: The value to look for.
    :type value: str
    :param default: Value to return if not found.
    :type default: any
    :returns: The object if it exists, otherwise the default value.
    :rtype: dict
    """
    return next((o for o in obj if key in o and o[key] == value), default)


def copy_field_values(src_app, src_field_name, dest_app, dest_field_name):
    """
    Copy a field's values from one field to another, only copying values
    that don't exist in the destination field. A best effort to preserve value
    order is made between the source and destination.

    :param src_app: The source app to copy from.
    :type src_app: :class:`App`
    :param src_field_name: the name of the field to copy from.
    :type src_field_name: str
    :param dest_app: The app to copy to.
    :type dest_app: :class:`App`
    :param dest_field_name: The name of the field to copy to.
    :type dest_field_name: str
    :returns: A tuple of added and moved values.
    :rtype: tuple
    """
    if not isinstance(src_app, App):
        raise TypeError('The src_app must be an instance of App')
    if not isinstance(dest_app, App):
        raise TypeError('The dest_app must be an instance of App')
    src_field = get_by_key_value(src_app.fields, 'name', src_field_name)
    if not src_field:
        raise ValueError('Field %s does not exist in source app' %
                         src_field_name)
    dest_field = get_by_key_value(dest_app.fields, 'name', dest_field_name)
    if not dest_field:
        raise ValueError('Field %s does not exist in destination app' %
                         dest_field_name)
    if src_field['fieldType'] != 'valuesList' or \
            dest_field['fieldType'] != 'valuesList':
        raise TypeError('Source and destination fields must be valuesList')
    added_values = []
    moved_values= []
    for src_index, src_value in enumerate(src_field['values']):
        dest_value = get_by_key_value(dest_field['values'], 'name',
                                      src_value['name'])
        if not dest_value:
            dest_value = src_value.copy()
            dest_value['id'] = random_objectid()
            dest_field['values'].insert(src_index, dest_value)
            added_values.append(dest_value)
        else:
            dest_index = dest_field['values'].index(dest_value)
            if src_index != dest_index:
                dest_value = dest_field['values'].pop(dest_index)
                dest_field['values'].insert(src_index, dest_value)
                moved_values.append(dest_value)
    if added_values or moved_values:
        dest_app.save()
    return added_values, moved_values



"""This module provides a Group class."""

from ..auth import Client
from .resource import Resource


class Group(Resource):
    """A class for working with Swimlane groups."""

    def __init__(self, fields):
        """Init a Group with fields.

        Args:
            fields (dict): A dict of fields and values.
        """
        super(Group, self).__init__(fields)

    @classmethod
    def find_all(cls):
        """Get all groups.

        Returns:
            A generator that yields all groups in the system.
        """
        return (Group(g) for g in Client.get("groups").get('groups', []))

    @classmethod
    def find(cls, group_id=None, name=None):
        """Find groups.

        Args:
            group_id (str): The group ID.
            name (str): A name or a fragment of a name to search by.

        Returns:
            A single Group if group_id was specified, otherwise a generator
            that yields Groups for all system groups with matching names..
        """
        if group_id:
            return Group(Client.get("groups/{0}".format(group_id)))

        return (Group(g) for g
                in Client.get("groups/lookup?name={0}".format(name)))



"""This module provides a client for working with Swimlane."""

from __future__ import absolute_import

import json
try:
    from urllib.parse import urljoin
except ImportError:
    from urlparse import urljoin
import requests
from combomethod import combomethod
from .. import SwimlaneDict
from .user_pass_auth_provider import UserPassAuthProvider

__metaclass__ = type


class Client():
    """A client for Swimlane's REST API."""

    # The default client used for all requests.
    default = None

    @classmethod
    def set_default(cls, base_url, username, password, verify_ssl=True):
        """Initialize a client that can make REST requests to Swimlane.

        Args:
            base_url (str): The base URL for Swimlane
            username (str): A Swimlane username
            password (str): The Swimlane user's password
            verify_ssl (bool): Whether or not to verify SSL certs when calling
                Swimlane (default is True).

        Returns:
            Client: An instance of a Client
        """
        cls.default = Client(base_url, username, password, verify_ssl)
        cls.default.connect()

    @classmethod
    def check_default_is_set(cls):
        """Check to see if Client.default has been set."""
        if not cls.default:
            raise Exception("There is no default client set. Call set_default "
                            "to establish a default client.")

    def __init__(self, base_url, username, password, verify_ssl=True):
        """Initialize a client that can make REST requests to Swimlane.

        Args:
            base_url (str): The base URL for Swimlane
            username (str): A Swimlane username
            password (str): The Swimlane user's password
            verify_ssl (bool): Whether or not to verify SSL certs when calling
                Swimlane (default is True).

        Returns:
            Client: An instance of a Client
        """
        self.headers = {}
        self.username = username
        self.verify_ssl = verify_ssl
        self.base_url = urljoin(base_url, "api/")
        self.provider = UserPassAuthProvider(self.base_url, username, password,
                                             self.verify_ssl)

    def connect(self):
        """Connect to the Swimlane server."""
        self.headers.update(self.provider.auth_header())

    @combomethod
    def get(self, url):
        """Get resource JSON as a dict.

        Args:
            url (str): The resource URL fragment.

        Returns
            dict: The resource JSON as a SwimlaneDict.
        """
        if self is Client:
            cls = self
            cls.check_default_is_set()
            return cls.default.get(url)

        full_url = urljoin(self.base_url, url)
        resp = requests.get(full_url, headers=self.headers,
                            verify=self.verify_ssl)
        resp.raise_for_status()
        return self.build_payload(resp)

    @combomethod
    def post(self, resource, url):
        """Post a resource to Swimlane.

        Args:
            resource (Resource): The resource.
            url (str): The url to POST to.

        Returns:
            OrderedDict: The response JSON as a SwimlaneDict.
        """
        if self is Client:
            cls = self
            cls.check_default_is_set()
            return cls.default.post(resource, url)

        return self.send_data(resource, url, requests.post)

    @combomethod
    def put(self, resource, url):
        """Put a resource to Swimlane.

        Args:
            resource (Resource): The resource.
            url (str): The url to PUT to.

        Returns:
            OrderedDict: The response JSON as a SwimlaneDict.
        """
        if self is Client:
            cls = self
            cls.check_default_is_set()
            return cls.default.put(resource, url)

        return self.send_data(resource, url, requests.put)

    def send_data(self, resource, url, func):
        """Send data to Swimlane.

        Args:
            resource (Resource): The resource.
            url (str): The url.
            func (callable): The requests function to use to send the data.

        Returns:
            dict: The response JSON as a SwimlaneDict.
        """
        full_url = urljoin(self.base_url, url)
        self.headers.update({"content-type": "application/json"})
        data = resource._fields if hasattr(resource, "_fields") else resource
        resp = func(full_url, data=json.dumps(data), headers=self.headers,
                    verify=self.verify_ssl)
        resp.raise_for_status()
        return self.build_payload(resp)

    def build_payload(self, resp):
        """Build a SwimlaneDict or a list of SwimlaneDicts from the response.

        Args:
            resp (requests.Response): The response.
        """
        if not resp.content:
            return None
        payload = resp.json()
        if isinstance(payload, list):
            return (SwimlaneDict(item) for item in payload)
        if isinstance(payload, dict):
            return SwimlaneDict(payload)
        return payload



import mock
import unittest

from swimlane.core.resources import Task


MOCK_TASK = {
    'id': '123',
    'name': 'Mock Task'
}


class TaskTestCase(unittest.TestCase):
    def test_init(self):
        task = Task(MOCK_TASK)
        for key, value in MOCK_TASK.items():
            self.assertEqual(getattr(task, key), value)

    @mock.patch('swimlane.core.resources.task.Client', autospec=True)
    def test_run(self, mock_client):
        mock_record = mock.MagicMock()
        mock_record._fields = {'foo': 'bar'}
        task = Task(MOCK_TASK)
        task.run(mock_record)
        mock_client.post.assert_called_once_with({
            'taskId': task.id,
            'record': mock_record._fields},
            'task/execute/record')

    @mock.patch('swimlane.core.resources.task.Client', autospec=True)
    def test_find_all(self, mock_client):
        Task.find_all()
        mock_client.get.assert_called_once_with('task/light')

    @mock.patch('swimlane.core.resources.task.Client', autospec=True)
    def test_find(self, mock_client):
        Task.find('Mock Task')
        mock_client.get.assert_called_once_with('task/light')



import unittest

from swimlane.core import SwimlaneDict


D = {
    'one': 1,
    'two': 2,
    'values': {
        'four': 4,
        'five': 5,
        '$type': 'GARBAGE',
        'six': {
            'seven': 7,
            '$type': 'MOREGARBAGE',
            'eight': 8,
            'nine': {
                'ten': 10,
                '$type': 'EVENMOREGARBAGE'
            }
        }
    },
    '$type': 'TOPLEVELGARBAGE'
}


class SwimlaneDictTestCase(unittest.TestCase):
    def test_swimlane_dict(self):
        sd = SwimlaneDict(D)
        self.assertEqual(('$type', 'TOPLEVELGARBAGE'), sd.popitem(last=False))
        self.assertEqual(
            ('$type', 'GARBAGE'),
            sd['values'].popitem(last=False))
        self.assertEqual(
            ('$type', 'MOREGARBAGE'),
            sd['values']['six'].popitem(last=False))
        self.assertEqual(
            ('$type', 'EVENMOREGARBAGE'),
            sd['values']['six']['nine'].popitem(last=False))



"""This module provides operators and functions for filtering."""

EQ = "equals"
NOT_EQ = "doesNotEqual"
CONTAINS = "contains"
EXCLUDES = "excludes"


def create_filter(field_id, filter_type, value):
    """Create a Filter.

    Args:
        field_id (str): The field ID.
        filter_type (str): The type of filter.
        value (str): The value to filter on.

    Returns:
        A dict representing the filter.
    """
    return {
        "fieldId": field_id,
        "filterType": filter_type,
        "value": value,
    }



import mock
import unittest

from swimlane.core.resources import App


MOCK_APP = {
    'id': '123',
    'name': 'Mock App',
    'acronym': 'MA',
    'fields': [{
        'id': '456',
        'name': 'Mock Field'
    }]
}


class AppTestCase(unittest.TestCase):
    def test_init(self):
        app = App(MOCK_APP)
        for key, value in MOCK_APP.items():
            self.assertEqual(getattr(app, key), value)

    def test_field_id(self):
        app = App(MOCK_APP)
        self.assertEqual(app.field_id('Mock Field'), '456')
        self.assertEqual(app.field_id('Not A Field'), None)

    @mock.patch('swimlane.core.resources.app.Client', autospec=True)
    def test_find_all(self, mock_client):
        mock_client.get.return_value = [MOCK_APP]
        apps = list(App.find_all())
        self.assertEqual(len(apps), 1)
        self.assertIsInstance(apps[0], App)

    @mock.patch('swimlane.core.resources.app.Client', autospec=True)
    def test_find_by_id(self, mock_client):
        mock_client.get.return_value = MOCK_APP
        app = App.find(app_id='123')
        self.assertIsInstance(app, App)
        self.assertEqual(app.id, '123')

    @mock.patch('swimlane.core.resources.app.Client', autospec=True)
    def test_find_by_name(self, mock_client):
        mock_client.get.return_value = [MOCK_APP]
        app = App.find(name='Mock App')
        self.assertIsInstance(app, App)
        self.assertEqual(app.name, 'Mock App')

    @mock.patch('swimlane.core.resources.app.Client', autospec=True)
    def test_find_by_acronym(self, mock_client):
        mock_client.get.return_value = [MOCK_APP]
        app = App.find(acronym='MA')
        self.assertIsInstance(app, App)
        self.assertEqual(app.acronym, 'MA')

    @mock.patch('swimlane.core.resources.app.Client', autospec=True)
    def test_find_by_name_does_not_exist(self, mock_client):
        mock_client.get.return_value = [MOCK_APP]
        app = App.find(name='Some Other App')
        self.assertIsNone(app)

    @mock.patch('swimlane.core.resources.app.Client', autospec=True)
    def test_find_by_acronym_does_not_exist(self, mock_client):
        mock_client.get.return_value = [MOCK_APP]
        app = App.find(acronym='SOA')
        self.assertIsNone(app)



"""This module provides operators and functions for aggregating."""

AVG = "average"
COUNT = "count"
SUM = "sum"
MIN = "min"
MAX = "max"


def create_aggregate(field_id, aggregate_type):
    """Create an aggregate..

    Args:
        field_id (str): The field ID.
        aggregate_type (str): The type of aggregate..

    Returns:
        A dict representing the aggregate..
    """
    return {
        "fieldId": field_id,
        "aggregateType": aggregate_type
    }




import mock
import unittest

from swimlane.core.resources import User


MOCK_USER = {
    'id': '123',
    'name': 'Mock User'
}

MOCK_USERS = {
    'users': [MOCK_USER]
}


class UserTestCase(unittest.TestCase):
    def test_init(self):
        user = User(MOCK_USER)
        for key, value in MOCK_USER.items():
            self.assertEqual(getattr(user, key), value)

    @mock.patch('swimlane.core.resources.user.Client', autospec=True)
    def test_find_all(self, mock_client):
        mock_client.get.return_value = MOCK_USERS
        users = list(User.find_all())
        self.assertEqual(len(users), 1)
        self.assertIsInstance(users[0], User)

    @mock.patch('swimlane.core.resources.user.Client', autospec=True)
    def test_find_by_id(self, mock_client):
        mock_client.get.return_value = MOCK_USER
        user = User.find(user_id='123')
        self.assertIsInstance(user, User)
        self.assertEqual(user.id, '123')

    @mock.patch('swimlane.core.resources.user.Client', autospec=True)
    def test_find_by_name(self, mock_client):
        mock_client.get.return_value = MOCK_USERS
        users = list(User.find(name='Mock User'))
        self.assertEqual(len(users), 1)
        self.assertIsInstance(users[0], User)

    @mock.patch('swimlane.core.resources.user.Client', autospec=True)
    def test_find_by_name_does_not_exist(self, mock_client):
        mock_client.get.return_value = []
        users = User.find(name='Some Other User')
        self.assertEqual(list(users), [])



"""This module provides functions that are useful for working with records."""

from .core.auth import Client
from .core.resources import App, Record, Report, User
from .core.search import Search


def add_references(record_id,
                   keywords,
                   app_id=None,
                   app_name=None,
                   remote_app_id=None,
                   remote_app_name=None,
                   field_id=None,
                   field_name=None):
    """Add references to a record by querying with keywords.

    This function creates references on a record in one app to records in
    another app. Note that you can specify the apps by ID, name, or acronym and
    you can specify the reference field by ID or name.

    Args:
        record_id (str): The ID of the record references will be added to.
        keywords (str): The keywords that will be used to search for records.
        app_id (str): An App ID.
        app_name (str): An App name.
        remote_app_id (str): The ID of the App to query against.
        remote_app_name (str): The name of the App to query against.
        field_id (str): The field ID that references will be added to.
        field_name (str): The field name that references will be added to.

    Returns:
        A list of the Records that were referenced.
    """
    app = App.find(app_id=app_id, name=app_name)

    if not app:
        raise Exception("Unable to find App")

    if not field_id and field_name:
        field_id = app.field_id(field_name)
    if not field_id:
        raise Exception("field_id or field_name must be supplied")

    remote_app = App.find(app_id=remote_app_id, name=remote_app_name)

    if not remote_app:
        raise Exception("Unable to find remote App")

    current_user = next(User.find(name=Client.default.username), None)

    report = Report.new_for(remote_app.id, current_user.id, "addrefs")
    report.keywords = keywords

    records = Search(report).execute().records

    if records:
        record = Record.find(app.id, record_id)
        if not record:
            raise Exception("Unable to find record")
        record.values[field_id] = [r.id for r in records]
        record.update()
        return record




import unittest

from swimlane.core.resources import StatsResult


MOCK_STATS_RESULT = {
    'foo': 'bar'
}


class StatsResultTestCase(unittest.TestCase):
    def test_init(self):
        stats_result = StatsResult(MOCK_STATS_RESULT)
        for key, value in MOCK_STATS_RESULT.items():
            self.assertEqual(getattr(stats_result, key), value)



import mock
import unittest

from swimlane.core.resources import StatsReport, User


MOCK_STATS_REPORT = {
    'id': '123',
    'name': 'Mock Stats Report'
}


MOCK_USER = User(fields={
    'id': '123',
    'name': 'Mock User'
})


class StatsReportTestCase(unittest.TestCase):
    def test_init(self):
        stats_report = StatsReport(MOCK_STATS_REPORT)
        for key, value in MOCK_STATS_REPORT.items():
            self.assertEqual(getattr(stats_report, key), value)

    @mock.patch('swimlane.core.resources.report.User', autospec=True)
    def test_new_for(self, mock_user):
        mock_user.find.return_value = MOCK_USER
        stats_report = StatsReport.new_for('123', '456', 'Test Report')
        self.assertIsInstance(stats_report, StatsReport)

    @mock.patch('swimlane.core.resources.report.Client', autospec=True)
    def test_find_all(self, mock_client):
        StatsReport.find_all()
        mock_client.get.assert_called_with('reports')
        StatsReport.find_all('123')
        mock_client.get.assert_called_with('reports?appId=123')



"""Provides a StatsResult class."""

from .resource import Resource


class StatsResult(Resource):
    """A simple abstraction over a Swimlane stats collection."""

    def __init__(self, fields):
        """Init a StatsResult with fields.

        Args:
            fields (dict): A dict of fields and values
        """
        super(StatsResult, self).__init__(fields)
        self.groups = self._fields




"""This module provides operators and functions for grouping by a field.."""

GB = "groupBy"
HOUR = "groupByHour"
DAY = "groupByDay"
WEEK = "groupByWeek"
MONTH = "groupByMonth"
QUARTER = "groupByQuarter"
YEAR = "groupByYear"


def create_groupby(field_id, groupby_type):
    """Create a GroupBy.

    Args:
        field_id (str): The field ID.
        groupby_type (str): The type of grouping.

    Returns:
        A dict representing the groupby.
    """
    return {
        "fieldId": field_id,
        "groupByType": groupby_type
    }




"""Provides a Record class."""

from datetime import datetime
from warnings import warn

from ..auth import Client
from .group import Group
from .resource import Resource
from .user import User


class Record(Resource):
    """A simple abstraction over a Swimlane record resource."""

    def __init__(self, fields):
        """Init a Record with fields.

        Args:
            fields (dict): A dict of fields and values
        """
        super(Record, self).__init__(fields)

    def insert(self):
        """Insert the current record."""
        warn(
            'Record.insert method will be deleted in v0.1.0, use Record.save',
            category=DeprecationWarning)
        self.save()

    def update(self):
        """Update the current record."""
        warn(
            'Record.update method will be deleted in v0.1.0, use Record.save',
            category=DeprecationWarning)
        self.save()

    def save(self):
        """Create/update a record."""
        if not hasattr(self, 'isNew') or self.isNew is True:
            self._fields = Client.post(
                self, "app/{0}/record".format(self.applicationId))
        else:
            self._fields = Client.put(
                self, "app/{0}/record".format(self.applicationId))

    def reload(self):
        """Reload a record instance."""
        if not hasattr(self, 'isNew') or self.isNew is True:
            raise Exception(
                'Cannot reload an unsaved record, call save() first')
        r = Record.find(self.applicationId, self.id)
        self._fields.update(**r._fields)

    def add_comment(self, field_id, user_id, message):
        """Add a comment to a field.

        Args:
            field_id (str): The field ID.
            user_id (str): The user ID of the commenting user.
            message (str): The comment message.
        """
        Client.post({
            "message": message,
            "createdDate": datetime.utcnow().isoformat() + "Z"
        }, "app/{0}/record/{1}/{2}/comment".format(
            self.applicationId, self.id, field_id))

    def references(self, field_id, record_ids, ref_field_ids):
        """Get referenced Records.

        Args:
            field_id (str): The ID of the field on the current record.
            record_ids (list): The IDs of any records to retrieve.
            field_ids (list): The IDs of any fields to retrieve.

        Returns:
            A generator that yields all referenced Records.
        """
        url = ("app/{0}/record/{1}/references"
               "?recordIds={2}&fieldIds={3}".format(
                   self.applicationId, self.id,
                   ",".join(record_ids),
                   ",".join(ref_field_ids)))
        return (Record(r) for r in Client.get(url))

    @classmethod
    def new_for(cls, app_id):
        """Get a prefilled Record for the App designated by app_id.

        Args:
            app_id (str): A valid App ID.

        Return:
            A dict containing default fields and values for a Record.
        """
        return Record(Client.get("app/{0}/record".format(app_id)))

    @classmethod
    def find(cls, app_id, record_id):
        """Find a Record by app_id and record_id.

        Args:
            app_id (str): A valid App ID
            record_id (str): A valid Record ID

        Return:
            A Record
        """
        return Record(Client.get("app/{0}/record/{1}".format(app_id,
                                                             record_id)))

    def restrict(self, *user_groups, **kwargs):
        """
        Restrict the record to specific users and/or groups.

        :param user_groups: One or more User/Group instances.
        :type user_groups: :class:`Group` or :class:`User`
        :param append: Append users/groups to existing restriction.
        :type append: bool
        """
        allowed = [ug.summary for ug in user_groups
                   if isinstance(ug, (Group, User))]
        if 'append' in kwargs and kwargs['append']:
            allowed = self.allowed + allowed
        Client.put(allowed, 'app/{0}/record/{1}/restrict'.format(
            self.applicationId, self.id))
        return allowed



import mock
import unittest

from swimlane.core.resources import Report, User
from swimlane.core.search.filtering import create_filter, EQ
MOCK_REPORT = {
    'id': '123',
    'name': 'Mock Report'
}


MOCK_USER = User(fields={
    'id': '123',
    'name': 'Mock User'
})


class ReportTestCase(unittest.TestCase):
    def test_init(self):
        report = Report(MOCK_REPORT)
        for key, value in MOCK_REPORT.items():
            self.assertEqual(getattr(report, key), value)

    @mock.patch('swimlane.core.resources.report.Client', autospec=True)
    def test_insert(self, mock_client):
        report = Report(MOCK_REPORT)
        report.insert()
        mock_client.post.assert_called_once_with(report, 'reports')

    @mock.patch('swimlane.core.resources.report.Client', autospec=True)
    def test_update(self, mock_client):
        report = Report(MOCK_REPORT)
        report.update()
        mock_client.put.assert_called_once_with(report, 'reports')

    @mock.patch('swimlane.core.resources.report.User', autospec=True)
    def test_new_for(self, mock_user):
        mock_user.find.return_value = MOCK_USER
        report = Report.new_for('123', '456', 'New Report')
        self.assertIsInstance(report, Report)

    @mock.patch('swimlane.core.resources.report.Client', autospec=True)
    def test_find_all(self, mock_client):
        Report.find_all()
        mock_client.get.assert_called_with('reports')
        Report.find_all('123')
        mock_client.get.assert_called_with('reports?appId=123')

    @mock.patch('swimlane.core.resources.report.Client', autospec=True)
    def test_find(self, mock_client):
        Report.find('123')
        mock_client.get.assert_called_once_with('reports/123')



import mock
import unittest

from swimlane.core.auth import Client
from swimlane.core.auth.user_pass_auth_provider import UserPassAuthProvider
from swimlane.core.swimlane_dict import SwimlaneDict


class ClientTestCase(unittest.TestCase):
    def test_init(self):
        c = Client('server', 'user', 'pass')

        self.assertEqual(c.headers, {})
        self.assertEqual(c.username, 'user')
        self.assertEqual(c.verify_ssl, True)
        self.assertEqual(c.base_url, 'api/')
        self.assertIsInstance(c.provider, UserPassAuthProvider)

    @mock.patch('swimlane.core.auth.client.UserPassAuthProvider', autospec=True)
    def test_connect(self, mock_provider):
        mock_cookie = {'Cookie': 'cookie=testcookie'}
        mock_provider.return_value.auth_header.return_value = mock_cookie

        c = Client('server', 'user', 'pass')
        c.connect()

        self.assertEqual(c.headers, mock_cookie)

    @mock.patch('swimlane.core.auth.client.UserPassAuthProvider', autospec=True)
    def test_set_default(self, mock_provider):
        Client.set_default('server', 'user', 'pass')
        mock_provider.return_value.auth_header.assert_called_once_with()

    def test_check_default_is_set(self):
        self.assertRaises(Exception, Client.check_default_is_set)

    @mock.patch('swimlane.core.auth.client.requests', autospec=True)
    def test_get(self, mock_requests):
        mock_response = mock.MagicMock()
        mock_requests.get.return_value = mock_response

        c = Client('server', 'user', 'pass')
        c.get('test')

        mock_requests.get.assert_called_once_with(
            'api/test',
            headers={},
            verify=True)
        mock_response.raise_for_status.assert_called_once_with()

    @mock.patch('swimlane.core.auth.client.requests', autospec=True)
    def test_get_classmethod(self, mock_requests):
        mock_response = mock.MagicMock()
        mock_requests.get.return_value = mock_response

        Client.default = Client('server', 'user', 'pass')
        Client.get('test')

        mock_requests.get.assert_called_once_with(
            'api/test',
            headers={},
            verify=True)
        mock_response.raise_for_status.assert_called_once_with()

    @mock.patch('swimlane.core.auth.client.requests', autospec=True)
    def test_post(self, mock_requests):
        mock_response = mock.MagicMock()
        mock_requests.post.return_value = mock_response

        c = Client('server', 'user', 'pass')
        c.post({}, 'test')

        mock_requests.post.assert_called_once_with(
            'api/test',
            data='{}',
            headers={'content-type': 'application/json'},
            verify=True)
        mock_response.raise_for_status.assert_called_once_with()

    @mock.patch('swimlane.core.auth.client.requests', autospec=True)
    def test_post(self, mock_requests):
        mock_response = mock.MagicMock()
        mock_requests.post.return_value = mock_response

        Client.default = Client('server', 'user', 'pass')
        Client.post({}, 'test')

        mock_requests.post.assert_called_once_with(
            'api/test',
            data='{}',
            headers={'content-type': 'application/json'},
            verify=True)
        mock_response.raise_for_status.assert_called_once_with()


    @mock.patch('swimlane.core.auth.client.requests', autospec=True)
    def test_put(self, mock_requests):
        mock_response = mock.MagicMock()
        mock_requests.put.return_value = mock_response

        c = Client('server', 'user', 'pass')
        c.put({}, 'test')

        mock_requests.put.assert_called_once_with(
            'api/test',
            data='{}',
            headers={'content-type': 'application/json'},
            verify=True)
        mock_response.raise_for_status.assert_called_once_with()

    @mock.patch('swimlane.core.auth.client.requests', autospec=True)
    def test_put_classmethod(self, mock_requests):
        mock_response = mock.MagicMock()
        mock_requests.put.return_value = mock_response

        Client.default = Client('server', 'user', 'pass')
        Client.put({}, 'test')

        mock_requests.put.assert_called_once_with(
            'api/test',
            data='{}',
            headers={'content-type': 'application/json'},
            verify=True)
        mock_response.raise_for_status.assert_called_once_with()


    def test_send_data(self):
        mock_requests = mock.MagicMock()

        c = Client('server', 'user', 'pass')
        c.send_data({}, 'test', mock_requests)

        mock_requests.assert_called_once_with(
            'api/test',
            data='{}',
            headers={'content-type': 'application/json'},
            verify=True)
        mock_requests.return_value.raise_for_status.assert_called_once_with()

    def test_build_payload(self):
        mock_response = mock.MagicMock()
        c = Client('server', 'user', 'pass')

        mock_response.content = None
        self.assertEqual(c.build_payload(mock_response), None)

        mock_response.content = '{}'
        mock_response.json.return_value = {}
        self.assertIsInstance(c.build_payload(mock_response), SwimlaneDict)

        mock_response.content = '[{}]'
        mock_response.json.return_value = [{}]
        payload = c.build_payload(mock_response)
        self.assertIsInstance(next(payload), SwimlaneDict)



from setuptools import setup, find_packages

setup(
    name="swimlane",
    author="Swimlane LLC",
    author_email="info@swimlane.com",
    version="0.1.0",
    url="https://github.com/Swimlane/sw-python-client",
    packages=find_packages(),
    description="A Python client for Swimlane.",
    install_requires=[
        "requests==2.8.1",
        "ordereddict==1.1",
        "combomethod==1.0.6"
    ],
    test_suite='nose.collector',
    tests_require=['coverage', 'mock', 'nose'],
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Development Status :: 4 - Beta",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: Implementation :: PyPy"
    ]
)



"""This module provides a Task class."""

from ..auth import Client
from .resource import Resource


class Task(Resource):
    """A simple abstraction of a Swimlane task."""

    def __init__(self, fields):
        """Init a Task with fields.

        Args:
            fields (dict): A dict of fields and values.
        """
        super(Task, self).__init__(fields)

    def run(self, record):
        """Run a task.

        Args:
            record (Record): The record to run this task against.
        """
        Client.post({
            "taskId": self.id,
            "record": record._fields
        }, "task/execute/record")

    @classmethod
    def find_all(cls):
        """Find all Tasks in the system.

        Returns:
            A generator that yields Tasks.
        """
        return (Task(t) for t in Client.get("task/light"))

    @classmethod
    def find(cls, name):
        """Find all Tasks in the system.

        Args:
            name (str): The task name.

        Returns:
            A Task.
        """
        return next((t for t in cls.find_all() if t.name == name), None)




"""This module provides a StatsReport class."""

from .. import SwimlaneDict
from .report import Report

TEMPLATE = list({
    "$type": "Core.Models.Search.StatsReport, Core",
    "chartOptions": {
        "showLegend": True,
        "showLabels": True,
        "showXAxis": True,
        "showYAxis": True,
        "showXAxisLabel": False,
        "showYAxisLabel": False,
        "xAxisLabelText": "",
        "yAxisLabelText": "",
        "sort": {
            "directionD0": "labelAscending",
            "directionD1": "labelAscending"
        },
        "zoom": False,
        "explodeSlices": False,
        "showOtherGroup": True,
        "chartType": "verticalBar",
        "colorScheme": "flame"
    },
    "statsDrillin": False
}.items())


class StatsReport(Report):
    """A stats report class used for searching."""

    def __init__(self, fields):
        """Init a StatsReport."""
        self._fields = fields

    @classmethod
    def new_for(cls, app_id, user_id, name):
        """Get a prefilled StatsReport for the App designated by app_id.

        Args:
            app_id (str): The ID of the App to search in.
            user_id (str): The ID of the user creating the report.
            name (str): The name of the Report.

        Return:
            A prefilled StatsReport.
        """
        report = super(StatsReport, cls).new_for(app_id, user_id, name)
        fields = list(report._fields.items())
        return StatsReport(SwimlaneDict(dict(fields + TEMPLATE)))

    @classmethod
    def find_all(cls, app_id=None):
        """Find all stats reports.

        If app_id is specified, only reports that are a member of that App
        will be returned. By default, all reports in the system are returned.

        Args:
            app_id (str): An App ID.

        Returns:
            A generator that yields StatsReports.
        """
        parent_all = super(StatsReport, cls).find_all(app_id)
        return (StatsReport(r._fields) for r in parent_all
                if "StatsReport" in r._fields["$type"])




import mock
import unittest

from swimlane.core.resources import StatsReport
from swimlane.core.search import Search
from swimlane.core.search.search_result import SearchResult


class SearchTestCase(unittest.TestCase):
    def test_init(self):
        mock_report = mock.Mock()
        search = Search(mock_report)
        self.assertEqual(search.report, mock_report)
        self.assertFalse(search.has_more_pages)
        pass

    @mock.patch('swimlane.core.search.search.Client', autospec=True)
    def test_execute(self, mock_client):
        mock_response = {
            'count': '1',
            'limit': '1',
            'offset': '1',
            'results': []
        }
        mock_client.post.return_value = mock_response

        report = StatsReport({
            'applicationIds': [{}],
            'offset': 1,
            'pageSize': 1})
        search = Search(report)
        search_result = search.execute()

        mock_client.post.assert_called_once_with(report, 'search/stats')
        self.assertIsInstance(search_result, SearchResult)

        mock_client.post.return_value = (r for r in [mock_response])
        search.next_page()
        self.assertEqual(len(mock_client.post.mock_calls), 2)



import unittest

from swimlane.core.resources import Resource


class ResourceTestCase(unittest.TestCase):
    def test_init(self):
        r = Resource({'foo': 1, 'bar': True, 'baz': None})

        self.assertEqual(r.foo, 1)
        self.assertTrue(r.bar)
        self.assertIsNone(r.baz)
        self.assertRaises(AttributeError, lambda: r.doesnotexist)
        r.bar = False
        self.assertFalse(r.bar)
        self.assertIn('foo', str(r))



import mock
import unittest

from swimlane.core.resources import Group

MOCK_GROUP = {
    'id': '123',
    'name': 'Mock Group'
}

MOCK_GROUPS = {
    'groups': [MOCK_GROUP]
}


class GroupTestCase(unittest.TestCase):
    def test_init(self):
        group = Group(MOCK_GROUP)
        for key, value in MOCK_GROUP.items():
            self.assertEqual(getattr(group, key), value)

    @mock.patch('swimlane.core.resources.group.Client', autospec=True)
    def test_find_all(self, mock_client):
        mock_client.get.return_value = MOCK_GROUPS
        groups = list(Group.find_all())
        self.assertEqual(len(groups), 1)
        self.assertIsInstance(groups[0], Group)

    @mock.patch('swimlane.core.resources.group.Client', autospec=True)
    def test_find_by_id(self, mock_client):
        mock_client.get.return_value = MOCK_GROUP
        group = Group.find(group_id='123')
        self.assertIsInstance(group, Group)
        self.assertEqual(group.id, '123')

    @mock.patch('swimlane.core.resources.group.Client', autospec=True)
    def test_find_by_name(self, mock_client):
        mock_client.get.return_value = MOCK_GROUPS
        groups = list(Group.find(name='Mock Group'))
        self.assertEqual(len(groups), 1)
        self.assertIsInstance(groups[0], Group)

    @mock.patch('swimlane.core.resources.group.Client', autospec=True)
    def test_find_by_name_does_not_exist(self, mock_client):
        mock_client.get.return_value = []
        groups = Group.find(name='Some Other Group')
        self.assertEqual(list(groups), [])
